#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdint.h>
#include <assert.h>
#include <limits.h>

#include "common.h"
#include "macros.h"
#include "io.h"
#include "xmalloc.h"
#include "list.h"
#include "timeout.h"
#include "buffer.h"
#include "sock-unix.h"
#include "sock-inet.h"
#include "misc.h"
#include "hal-proto.h"
#include "hal-cmd-parser.h"
#include "hal-cmd.h"
#include "hal-mgr.h"
#include "hal-peer.h"

//#define DEBUG_HAL_MGR

#define CLT_TIMEOUT_MSEC	3000

#ifdef DEBUG_HAL_MGR
#define debug(fmt, arg...) fprintf(stderr, "debug: %s: "fmt CRLF, \
						__FUNCTION__,  ##arg);
#else
#define debug(fmt, arg...) do {} while (0)
#endif

#define warn(fmt, arg...) fprintf(stderr, "warn: %s: "fmt CRLF, \
						__FUNCTION__,  ##arg);

#define info(fmt, arg...) fprintf(stderr, "info: %s: "fmt CRLF, \
						__FUNCTION__,  ##arg);

/* Client states. */
enum {
	HAL_CLT_STATE_CONN = 0,
	HAL_CLT_STATE_DISCONN,
	HAL_CLT_STATE_ERROR,
	/* waiting reply states */
	HAL_CLT_STATE_CONN_WAIT,
	HAL_CLT_STATE_DISCONN_WAIT,
	HAL_CLT_STATE_RP_WAIT,
	HAL_CLT_STATE_WP_WAIT,
	HAL_CLT_STATE_RI_WAIT,
	HAL_CLT_STATE_RO_WAIT,
	HAL_CLT_STATE_WI_WAIT,
	HAL_CLT_STATE_WO_WAIT,
	HAL_CLT_STATE_RE_WAIT,
	HAL_CLT_STATE_WE_WAIT,
	HAL_CLT_STATE_CB_WAIT,
	HAL_CLT_STATE_CR_WAIT,
	HAL_CLT_STATE_CE_WAIT
};

/* Connection types. */
enum {
	HAL_CONN_TYPE_RAW,
	HAL_CONN_TYPE_UNIX,
	HAL_CONN_TYPE_INET
};

/* Structure represents a single connection. It is based on a peer. */
struct hal_client {
	struct hal_peer_base	base; 

	char			name[NAME_MAX];		/* client name */
	struct buffer		*inputs;		/* points to inputs_image */
	struct buffer		*outputs;		/* points to outputs_image */

	unsigned int		inputs_sz;		/* inputs client image size */
	unsigned int		inputs_offset;		/* offset in inputs_image */
	unsigned int		outputs_sz;		/* outputs client image size */
	unsigned int		outputs_offset;		/* offset in outputs_image */

	unsigned int		last_offset;		/* offset of last read/write operation */
	int			side;			/* client side */
	int			state;			/* client current state */
	struct timeout		*tout;			/* client timeout */
	int			response_timeout;

	int			image_pos;		/* position of slave or master
							   client in final images */
	int			conn_type;
	union {
		char		sock[PATH_MAX];		/* full path to a socket file */
		char		inet_addr[PATH_MAX];	/* hostname:port or ipaddr:port */
	};
};

/* Lists of clients. Lists are connected with client's non-waiting states. */
static struct slist		*clt_conn_list;
static struct slist		*clt_disconn_list;
static struct slist		*clt_error_list;

static struct io_context	*io;
static struct buffer		*inputs_image;
static struct buffer		*outputs_image;
static unsigned int		clt_num;		/* count masters and slaves only */

/* Move a client to a new list. */
static void client_move(struct slist **dst, struct hal_client *s)
{
	struct slist *list;

	list = slist_find(clt_conn_list, s, NULL);
	if (list != NULL) {
		clt_conn_list = slist_unlink(clt_conn_list, list);
		goto add;
	}

	list = slist_find(clt_disconn_list, s, NULL);
	if (list != NULL) {
		clt_disconn_list = slist_unlink(clt_disconn_list, list);
		goto add;
	}

	list = slist_find(clt_error_list, s, NULL);
	if (list != NULL) {
		clt_error_list = slist_unlink(clt_error_list, list);
		goto add;
	}

	assert(list != NULL);
add:

	*dst = slist_add_tail(*dst, list);
}

/* Common function for a set of commands. */
static void clients_cmd_common(const char *cmd, int side)
{
	struct hal_client *s;
	void (*request)(struct buffer *);
	struct slist *iter;
	int new_state, run;

	request = NULL;

	/* Depends on the command choose a new state and a request function. */
	if (cmd == HAL_CMD_READ_PREPARE) {
		new_state = HAL_CLT_STATE_RP_WAIT;
		request = hal_cmd_read_prepare_request;
	} else if (cmd == HAL_CMD_WRITE_PREPARE) {
		new_state = HAL_CLT_STATE_WP_WAIT;
		request = hal_cmd_write_prepare_request;
	} else if (cmd == HAL_CMD_READ_END) {
		new_state = HAL_CLT_STATE_RE_WAIT;
		request = hal_cmd_read_end_request;
	} else if (cmd == HAL_CMD_WRITE_END) {
		new_state = HAL_CLT_STATE_WE_WAIT;
		request = hal_cmd_write_end_request;
	}

	assert(request != NULL);
	run = 0;

	for (iter = clt_conn_list; iter != NULL; iter = iter->next) {
		s = (struct hal_client *)iter->data;

		assert(s->state == HAL_CLT_STATE_CONN);

		if (s->side != side)
			continue;
		/* Come to a new state and call our function to send a request. */
		s->state = new_state;
		request(HAL_PEER_BASE(s)->wb);

		debug("%s request", cmd);
		/* Send data and add timeout. */
		hal_peer_send_default(HAL_PEER_BASE(s));
		timeout_add(s->tout, s->response_timeout);
		run++;
	}

	/* Run a loop till not a loop stop condition. */
	if (run)
		io_loop_run(io);
}

/* Simple wrappers for a set of commands. */
static void clients_read_prepare(int side)
{
	clients_cmd_common(HAL_CMD_READ_PREPARE, side);
}

static void clients_write_prepare(int side)
{
	clients_cmd_common(HAL_CMD_WRITE_PREPARE, side);
}

static void clients_read_end(int side)
{
	clients_cmd_common(HAL_CMD_READ_END, side);
}

static void clients_write_end(int side)
{
	clients_cmd_common(HAL_CMD_WRITE_END, side);
}

/* Helpers for reading images (inputs and outputs). */
static void clients_read_image(const char *cmd, int side)
{
	struct hal_client *s;
	void (*read_request)(struct buffer *, unsigned int, unsigned int);
	struct slist *iter;
	unsigned int len, len_max;
	int new_state, run;

	read_request = NULL;
	
	/* Assign a new state and a request function. */
	if (cmd == HAL_CMD_READ_INPUTS) {
		new_state = HAL_CLT_STATE_RI_WAIT;
		read_request = hal_cmd_read_inputs_request;	
	} else if (cmd == HAL_CMD_READ_OUTPUTS) {
		new_state = HAL_CLT_STATE_RO_WAIT;
		read_request = hal_cmd_read_outputs_request;
	}

	assert(read_request != NULL);
	run = 0;
	/* Pass through all clients and call our request function. */
	for (iter = clt_conn_list; iter != NULL; iter = iter->next) {
		s = (struct hal_client *)iter->data;
		
		assert(s->state == HAL_CLT_STATE_CONN);

		if (s->side != side)
			continue;

		len_max = HAL_IMAGE_MAX >> 1;

		if (cmd == HAL_CMD_READ_INPUTS)
			len = s->inputs_sz > len_max ? len_max : s->inputs_sz;
		else if (cmd == HAL_CMD_READ_OUTPUTS) 
			len = s->outputs_sz > len_max ? len_max : s->outputs_sz;
		
		s->state = new_state;
		read_request(HAL_PEER_BASE(s)->wb, len, 0);
		debug("%s request: length %d offset 0", cmd, len);
		
		/* Keer track of the offset, to repeat request if the sent
		 * buffer size is not enough big. */
		s->last_offset = len;
		hal_peer_send_default(HAL_PEER_BASE(s));
		timeout_add(s->tout, s->response_timeout);
		run++;
	}
	
	if (run)
		io_loop_run(io);
}

static void clients_read_inputs(int side)
{
	clients_read_image(HAL_CMD_READ_INPUTS, side);
}

static void clients_read_outputs(int side)
{
	clients_read_image(HAL_CMD_READ_OUTPUTS, side);
}

static void image_fill(char image[HAL_IMAGE_MAX+1], struct buffer *src,
		       unsigned int off, unsigned int len)
{
	unsigned char *p;
	unsigned int i;

	p = buffer_ptr(src);
	p += off;

	for (i = 0; i < len; i++) {
		image[2*i] = hex_to_ascii((p[i] >> 4) & 0x0f);
		image[2*i+1] = hex_to_ascii(p[i] & 0x0f);
	}
	/* Terminate a string. */
	image[2*len] = 0;
}

static void buffer_fill(const char image[HAL_IMAGE_MAX+1], struct buffer *dst,
			unsigned int off)
{
	unsigned char *p, n;
	unsigned int len, i;

	len = strlen(image);
	assert(len % 2 == 0);

	p = buffer_ptr(dst);
	p += off;

	for (i = 0; i < len; i+= 2) {
		n = ascii_to_hex(image[i+1]);
		n += ascii_to_hex(image[i]) << 4;
		*p++ = n;
	}
}

/* Helpers for writing images (inputs and outputs). */
static void client_write_image(const char *cmd, int side)
{
	struct hal_client *s;
	void (*write_request)(struct buffer *, unsigned int, const char *);
	struct slist *iter;
	char image[HAL_IMAGE_MAX+1];
	unsigned int len, len_max;
	int new_state, run;

	write_request = NULL;
	
	if (cmd == HAL_CMD_WRITE_INPUTS) {
		new_state = HAL_CLT_STATE_WI_WAIT;
		write_request = hal_cmd_write_inputs_request;
	} else if (cmd == HAL_CMD_WRITE_OUTPUTS) {
		new_state = HAL_CLT_STATE_WO_WAIT;
		write_request = hal_cmd_write_outputs_request;
	}

	assert(write_request != NULL);
	run = 0;

	/* Pass through all clients and call our request function. */
	for (iter = clt_conn_list; iter != NULL; iter = iter->next) {
		s = (struct hal_client *)iter->data;

		assert(s->state == HAL_CLT_STATE_CONN);

		if (s->side != side)
			continue;

		s->state = new_state;
		len_max = HAL_IMAGE_MAX >> 1;

		/* Fill image buffer with data from inputs or outputs storage. */
		if (cmd == HAL_CMD_WRITE_INPUTS) {
			len = s->inputs_sz > len_max ? len_max : s->inputs_sz;
			image_fill(image, s->inputs, s->inputs_offset, len);
		} else if (cmd == HAL_CMD_WRITE_OUTPUTS) {
			len = s->outputs_sz > len_max ? len_max : s->outputs_sz;
			image_fill(image, s->outputs, s->outputs_offset, len);
		}
	
		write_request(HAL_PEER_BASE(s)->wb, 0, image);
		debug("%s request: length %d offset 0", cmd, len);

		s->last_offset = len;
		hal_peer_send_default(HAL_PEER_BASE(s));
		timeout_add(s->tout, s->response_timeout);
		run++;
	}
	
	if (run)
		io_loop_run(io);
}

static void clients_write_inputs(int side)
{
	client_write_image(HAL_CMD_WRITE_INPUTS, side);
}

static void clients_write_outputs(int side)
{
	client_write_image(HAL_CMD_WRITE_OUTPUTS, side);
}

/* Main functions. */

static void hal_mgr_clients_read_inputs()
{
	clients_read_prepare(HAL_CLT_SLAVE);
	clients_read_prepare(HAL_CLT_MASTER);
	clients_read_inputs(HAL_CLT_SLAVE);
	clients_read_end(HAL_CLT_SLAVE);
	clients_read_outputs(HAL_CLT_MASTER);
	clients_read_end(HAL_CLT_MASTER);
}

static void hal_mgr_clients_write_outputs()
{
	clients_write_prepare(HAL_CLT_SLAVE);
	clients_write_prepare(HAL_CLT_MASTER);
	clients_write_outputs(HAL_CLT_SLAVE);
	clients_write_end(HAL_CLT_SLAVE);
	clients_write_inputs(HAL_CLT_MASTER);
	clients_write_end(HAL_CLT_MASTER);
}

static void hal_mgr_runtime_write_images()
{
	clients_write_inputs(HAL_CLT_RUNTIME);
	clients_write_outputs(HAL_CLT_RUNTIME);
}

static void hal_mgr_runtime_read_images()
{
	clients_read_inputs(HAL_CLT_RUNTIME);
	clients_read_outputs(HAL_CLT_RUNTIME);
}

static void hal_mgr_cycle_begin()
{
	hal_mgr_clients_read_inputs();
	hal_mgr_runtime_write_images();
}

static void hal_mgr_cycle_end()
{
	hal_mgr_runtime_read_images();
	hal_mgr_clients_write_outputs();
}

static int compare(const void *a, const void *b)
{
	int *c, *d;

	c = (int *)a;
	d = (int *)b;

	if (*c > *d)
		return 1;
	else if (*c < *d)
		return -1;

	return 0;
}

/* Physically we alredy connected to other side. Now connect with a hal logic.
 * When all clients respond to us we can form the final inputs and outputs
 * images.
 */
int hal_mgr_clients_connect()
{
	struct hal_client *s;
	struct slist *iter, *list;
	struct buffer *inputs, *outputs;
	int run, connected, i;

	run = 0;

	/* Send connect request command to mastes and slaves. */
	for (iter = clt_disconn_list; iter != NULL; iter = iter->next) {
		s = (struct hal_client *)iter->data;

		if (s->side == HAL_CLT_RUNTIME)
			continue;
		
		assert(s->state == HAL_CLT_STATE_DISCONN);

		s->state = HAL_CLT_STATE_CONN_WAIT;
		hal_cmd_connect_request(HAL_PEER_BASE(s)->wb);
		hal_peer_send_default(HAL_PEER_BASE(s));
		timeout_add(s->tout, s->response_timeout);
		run++;
	}

	if (run)
		io_loop_run(io);

	connected = 0;
	/* The loop is stopped, check that all masters and slaves are
	 * connected now. */
	for (iter = clt_conn_list; iter != NULL; iter = iter->next) {
		s = (struct hal_client *)iter->data;

		if (s->side == HAL_CLT_RUNTIME)
			continue;
		
		connected++;
	}

	/* It means we can't form the final image. */
	if (run != connected)
		return -1;

	for (i = 0; i < connected; i++) {
		list = slist_find(clt_conn_list, &i, compare);
		assert(list != NULL);
		s = (struct hal_client *)list->data;
		
		/* The important stage. Master inputs and outputs are reversed
		 * to hal-mgr and vmachined. */
		switch (s->side) {
		case HAL_CLT_SLAVE:
			inputs = inputs_image;
			outputs = outputs_image;
			break;
		case HAL_CLT_MASTER:
			inputs = outputs_image;
			outputs = inputs_image;
			break;
		}

		s->inputs = inputs;
		s->inputs_offset = buffer_size(inputs);
		buffer_advance(inputs, s->inputs_sz);

		s->outputs = outputs;
		s->outputs_offset = buffer_size(outputs);
		buffer_advance(outputs, s->outputs_sz);
	}

	return 0;
}

static void connect_request_cb(void *opaque)
{
	struct hal_client *s;

	s = (struct hal_client *)opaque;

	if (s->side != HAL_CLT_RUNTIME) {
		warn("CN request: not runtime side");
		hal_status_error_reply(HAL_PEER_BASE(s)->wb, HAL_CMD_CONNECT,
						"You are not a runtime");
		goto send;
	}

	if (s->state != HAL_CLT_STATE_DISCONN) {
		warn("CN request: unexpected state");
		hal_status_error_reply(HAL_PEER_BASE(s)->wb, HAL_CMD_CONNECT,
						"Unexpected state");
		goto send;
	}

	/* First come to the connected list and then check that the disconnected
	 * list is empty. */
	client_move(&clt_conn_list, s);
	s->state = HAL_CLT_STATE_CONN;
	
	if (clt_disconn_list != NULL) {
		warn("CN request: not all clients are connected");
		hal_status_error_reply(HAL_PEER_BASE(s)->wb, HAL_CMD_CONNECT,
						"Not all clients are connected");
		client_move(&clt_disconn_list, s);
		s->state = HAL_CLT_STATE_DISCONN;
		goto send;	
	}

	s->inputs_offset = 0;
	s->inputs_sz = buffer_size(inputs_image);
	s->inputs = inputs_image;
	s->outputs_offset = 0;
	s->outputs_sz = buffer_size(outputs_image);
	s->outputs = outputs_image;

	hal_cmd_connect_reply(HAL_PEER_BASE(s)->wb, s->inputs_sz, s->outputs_sz, 0);
	debug("CN request: runtume connected");
	debug("Final images: %d inputs size, %d outputs size",
				s->inputs_sz, s->outputs_sz);

send:
	hal_peer_send_default(HAL_PEER_BASE(s));
}

/* A handler for cycle request commands. */
static void common_request(const char *cmd, void *opaque)
{
	struct hal_client *s;

	s = (struct hal_client *)opaque;

	if (s->side != HAL_CLT_RUNTIME) {
		warn("%s request: not runtime side", cmd);
		hal_status_error_reply(HAL_PEER_BASE(s)->wb, cmd,
						"You are not a runtime");
		goto send;
	}

	if (s->state != HAL_CLT_STATE_CONN) {
		warn("%s request: unexpected state", cmd);
		hal_status_error_reply(HAL_PEER_BASE(s)->wb, cmd,
						"Runtime not in connected state");
		goto send;
	}

	if (cmd == HAL_CMD_CYCLE_BEGIN) {
		hal_mgr_cycle_begin();
		hal_cmd_cycle_begin_reply(HAL_PEER_BASE(s)->wb);
	} else if (cmd == HAL_CMD_CYCLE_RUN) {
		hal_cmd_cycle_run_reply(HAL_PEER_BASE(s)->wb);
	} else if (cmd == HAL_CMD_CYCLE_END) {
		hal_mgr_cycle_end();
		hal_cmd_cycle_end_reply(HAL_PEER_BASE(s)->wb);
	} else if (cmd == HAL_CMD_DISCONNECT) {
		hal_cmd_disconnect_reply(HAL_PEER_BASE(s)->wb);
	}

	debug("%s request: OK", cmd);

send:
	hal_peer_send_default(HAL_PEER_BASE(s));
}

static void disconnect_request_cb(void *opaque)
{
	common_request(HAL_CMD_DISCONNECT, opaque);
}

static void cycle_begin_request_cb(void *opaque)
{
	common_request(HAL_CMD_CYCLE_BEGIN, opaque);
}

static void cycle_run_request_cb(void *opaque)
{
	common_request(HAL_CMD_CYCLE_RUN, opaque);
}

static void cycle_end_request_cb(void *opaque)
{
	common_request(HAL_CMD_CYCLE_END, opaque);
}

static void connect_reply_cb(unsigned int inputs_sz, unsigned int outputs_sz,
			     char side, const char *error, void *opaque)
{
	struct hal_client *s;

	s = (struct hal_client *)opaque;

	if (timeout_active(s->tout))
		timeout_del(s->tout);

	if (error != NULL) {
		warn("CN reply: %s", error);
		s->state = HAL_CLT_STATE_ERROR;
		client_move(&clt_error_list, s);
		return;
	}

	if (s->state != HAL_CLT_STATE_CONN_WAIT) {
		warn("CN reply: unexpected state");
		s->state = HAL_CLT_STATE_ERROR;
		client_move(&clt_error_list, s);
		return;
	}

	/* Check a config specified side and a client side. */
	switch (side) {
	case 'm':
		if (s->side != HAL_CLT_MASTER) {
			warn("CN reply: client %s wrong side", s->name);
			return;
		}
		break;
	case 's':
		if (s->side != HAL_CLT_SLAVE) {
			warn("CN reply: client %s wrong side", s->name);
			return;
		}
		break;
	}

	s->inputs_sz = inputs_sz;
	s->outputs_sz = outputs_sz;

	s->state = HAL_CLT_STATE_CONN;
	client_move(&clt_conn_list, s);

	debug("CN reply: inputs size %d, outputs size %d",
					inputs_sz, outputs_sz);
}

static void rw_common_reply(const char *cmd, const char *error, void *opaque)
{
	struct hal_client *s;
	int cur_state;

	s = (struct hal_client *)opaque;

	if (timeout_active(s->tout))
		timeout_del(s->tout);

	if (error != NULL) {
		warn("%s reply: %s", cmd, error);
		s->state = HAL_CLT_STATE_ERROR;
		client_move(&clt_error_list, s);
		return;
	}

	if (cmd == HAL_CMD_READ_PREPARE)
		cur_state = HAL_CLT_STATE_RP_WAIT;
	else if (cmd == HAL_CMD_WRITE_PREPARE)
		cur_state = HAL_CLT_STATE_WP_WAIT;
	else if (cmd == HAL_CMD_READ_END)
		cur_state = HAL_CLT_STATE_RE_WAIT;
	else if (cmd == HAL_CMD_WRITE_END)
		cur_state = HAL_CLT_STATE_WE_WAIT;

	if (s->state != cur_state) {
		warn("%s reply: unexpected state", cmd);
		s->state = HAL_CLT_STATE_ERROR;
		client_move(&clt_error_list, s);
		return;
	}

	s->state = HAL_CLT_STATE_CONN;

	debug("%s reply", cmd);
}

static void read_prepare_reply_cb(const char *error, void *opaque)
{
	rw_common_reply(HAL_CMD_READ_PREPARE, error, opaque);
}

static void write_prepare_reply_cb(const char *error, void *opaque)
{
	rw_common_reply(HAL_CMD_WRITE_PREPARE, error, opaque);
}

static void read_end_reply_cb(const char *error, void *opaque)
{
	rw_common_reply(HAL_CMD_READ_END, error, opaque);
}

static void write_end_reply_cb(const char *error, void *opaque)
{
	rw_common_reply(HAL_CMD_WRITE_END, error, opaque);
}

/* Handler for read reply commands. */
static void read_image_reply(const char *cmd, const char image[HAL_IMAGE_MAX+1],
			     unsigned int off, const char *error, void *opaque)
{
	struct hal_client *s;
	unsigned int len_next, len_max, rest, image_sz;
	int cur_state;
	void (*read_request)(struct buffer *, unsigned int, unsigned int);

	s = (struct hal_client *)opaque;

	if (timeout_active(s->tout))
		timeout_del(s->tout);

	if (error != NULL) {
		warn("%s reply: %s", cmd,  error);
		s->state = HAL_CLT_STATE_ERROR;
		client_move(&clt_error_list, s);
		return;
	}
	
	read_request = NULL;
	/* Handler processes two commands and we check then against different states. */
	if (cmd == HAL_CMD_READ_INPUTS)
		cur_state = HAL_CLT_STATE_RI_WAIT;
	else if (cmd == HAL_CMD_READ_OUTPUTS)
		cur_state = HAL_CLT_STATE_RO_WAIT;

	if (s->state != cur_state) {
		warn("%s reply: unexpected state", cmd);
		s->state = HAL_CLT_STATE_ERROR;
		client_move(&clt_error_list, s);
		return;
	}
	
	/* Copy data to inputs or outputs storage from received image buffer. */
	if (cmd == HAL_CMD_READ_INPUTS) {
		buffer_fill(image, s->inputs, s->inputs_offset+off);
		image_sz = s->inputs_sz;
		read_request = hal_cmd_read_inputs_request;
	} else if (cmd == HAL_CMD_READ_OUTPUTS) {
		buffer_fill(image, s->outputs, s->outputs_offset+off);
		image_sz = s->outputs_sz;
		read_request = hal_cmd_read_outputs_request;
	}

	len_max = HAL_IMAGE_MAX >> 1;
	/* In this case we need to read more. Send request for reading. */
	if (s->last_offset < image_sz) {
		rest = image_sz - s->last_offset;
		len_next = rest > len_max ? len_max : rest;

		assert(read_request != NULL);
		read_request(HAL_PEER_BASE(s)->wb, len_next, s->last_offset);
		debug("%s request: length  %d offset %d", cmd, len_next, s->last_offset);
		s->last_offset += len_next;

		hal_peer_send_default(HAL_PEER_BASE(s));
		timeout_add(s->tout, s->response_timeout);
	} else {
		s->state = HAL_CLT_STATE_CONN;
		debug("%s reply: come to connected state", cmd);
	}

	debug("%s reply", cmd);
}

static void read_inputs_reply_cb(const char inputs[HAL_IMAGE_MAX+1], unsigned int off,
				 const char *error, void *opaque)
{
	read_image_reply(HAL_CMD_READ_INPUTS, inputs, off, error, opaque);
}

static void read_outputs_reply_cb(const char outputs[HAL_IMAGE_MAX+1], unsigned int off,
				  const char *error, void *opaque)
{
	read_image_reply(HAL_CMD_READ_OUTPUTS, outputs, off, error, opaque);
}

/* Handler for a write reply command. */
static void write_image_reply(const char *cmd, const char *error, void *opaque)
{
	struct hal_client *s;
	void (*write_request)(struct buffer *, unsigned int, const char *);
	unsigned int  len_next, len_max, rest, image_sz;
	int cur_state;
	char image[HAL_IMAGE_MAX+1];

	s = (struct hal_client *)opaque;

	if (timeout_active(s->tout))
		timeout_del(s->tout);
	
	if (error != NULL) {
		warn("%s reply: %s", cmd,  error);
		s->state = HAL_CLT_STATE_ERROR;
		client_move(&clt_error_list, s);
		return;
	}

	write_request = NULL;

	if (cmd == HAL_CMD_WRITE_INPUTS)
		cur_state = HAL_CLT_STATE_WI_WAIT;
	else if (cmd == HAL_CMD_WRITE_OUTPUTS)
		cur_state = HAL_CLT_STATE_WO_WAIT;

	if (s->state != cur_state) {
		warn("%s reply: unexpected state", cmd);
		s->state = HAL_CLT_STATE_ERROR;
		client_move(&clt_error_list, s);
		return;
	}

	if (cmd == HAL_CMD_WRITE_INPUTS) {
		image_sz = s->inputs_sz;	
		write_request = hal_cmd_write_inputs_request;
	} else if (cmd == HAL_CMD_WRITE_OUTPUTS) {
		image_sz = s->outputs_sz;
		write_request = hal_cmd_write_outputs_request;
	}

	len_max = HAL_IMAGE_MAX >> 1;

	if (s->last_offset < image_sz) {
		rest = image_sz - s->last_offset;
		len_next = rest > len_max ? len_max : rest;
		/* Image contains new portion of data. */
		if (cmd == HAL_CMD_WRITE_INPUTS)
			image_fill(image, s->inputs, s->inputs_offset+s->last_offset, len_next);
		else if (cmd == HAL_CMD_WRITE_OUTPUTS)
			image_fill(image, s->outputs, s->outputs_offset+s->last_offset, len_next);

		assert(write_request != NULL);
		write_request(HAL_PEER_BASE(s)->wb, s->last_offset, image);
		debug("%s request: length  %d offset %d", cmd, len_next, s->last_offset);
		s->last_offset += len_next;

		hal_peer_send_default(HAL_PEER_BASE(s));
		timeout_add(s->tout, s->response_timeout);
	} else {
		s->state = HAL_CLT_STATE_CONN;
		debug("%s reply: come to connected state", cmd);
	}

	debug("%s reply", cmd);
}

static void write_inputs_reply_cb(const char *error, void *opaque)
{
	write_image_reply(HAL_CMD_WRITE_INPUTS, error, opaque);
}

static void write_outputs_reply_cb(const char *error, void *opaque)
{
	write_image_reply(HAL_CMD_WRITE_OUTPUTS, error, opaque);
}

static void not_handled_cb()
{       
	warn("Get unexpected command");
}

/* Set up hal parser handlers. */
static void parser_init()
{
	struct hal_cmd_callbacks parser_callbacks;

	parser_callbacks.connect_reply_callback = connect_reply_cb;
	parser_callbacks.connect_request_callback = connect_request_cb;
	parser_callbacks.disconnect_request_callback = disconnect_request_cb;

	parser_callbacks.cycle_begin_request_callback = cycle_begin_request_cb;
	parser_callbacks.cycle_end_request_callback = cycle_end_request_cb;
	parser_callbacks.cycle_run_request_callback = cycle_run_request_cb;

	parser_callbacks.read_prepare_reply_callback = read_prepare_reply_cb;
	parser_callbacks.write_prepare_reply_callback = write_prepare_reply_cb;
	parser_callbacks.read_end_reply_callback = read_end_reply_cb;
	parser_callbacks.write_end_reply_callback = write_end_reply_cb;

	parser_callbacks.read_inputs_reply_callback = read_inputs_reply_cb;
	parser_callbacks.read_outputs_reply_callback = read_outputs_reply_cb;
	parser_callbacks.write_inputs_reply_callback = write_inputs_reply_cb;
	parser_callbacks.write_outputs_reply_callback = write_outputs_reply_cb;
	parser_callbacks.not_handled_callback = not_handled_cb;

	hal_cmd_parser_register_callbacks(&parser_callbacks);
}

static void client_disable(struct hal_client *s)
{
	struct slist *list;

	hal_peer_base_deinit(HAL_PEER_BASE(s));

	assert(s->state != HAL_CLT_STATE_ERROR);
	assert(s->state != HAL_CLT_STATE_DISCONN);

	/* Runtime is a special case, it connects to us so we don't move it to
	 * error state list, we just free peer resources. */
	switch (s->side) {
	case HAL_CLT_RUNTIME:
		warn("disable a runtime");
		list = slist_find(clt_conn_list, s, NULL);
		clt_conn_list = slist_unlink(clt_conn_list, list);
		if (timeout_active(s->tout))
			timeout_del(s->tout);
		timeout_free(&s->tout);
		xfree(s);
		slist_free(&list);
		break;
	default:
		debug("move peer to a error state list");
		s->state = HAL_CLT_STATE_ERROR;
		client_move(&clt_error_list, s);
		break;
	}
}

/* Base peer handler. We monitor states and stop a loop when we need. */
static void peer_event_handler(struct hal_peer_base *peer,
			       hal_peer_event_t event)
{
	struct hal_client *s;
	struct slist *iter;
	int in_progress;

	s = (struct hal_client *)peer;
	in_progress = 0;

	switch (event) {
	case HAL_PEER_EVENT_TIMEOUT:
		warn("got timeout event");
		client_disable(s);
		return;
	case HAL_PEER_EVENT_ERROR:
		warn("got error event");
		client_disable(s);
		return;
	default:
		break;
	}

	/* When all clients come back to connected state we stop a loop. */
	for (iter = clt_conn_list; iter != NULL; iter = iter->next) {
		s = (struct hal_client *)iter->data;

		if (s->state != HAL_CLT_STATE_CONN)
			in_progress++;
	}

	debug("%d in progress", in_progress);

	if (in_progress == 0)
		io_loop_stop(io);
}
	
static struct hal_client *client_new(const char *name, int conn_type,
				     int timeout, int side)
{
	struct hal_client *s;
	struct slist *list;

	/* Allocate memory for a new session.*/
	s = xmalloc0(sizeof(*s));

	if (timeout > 0)
		s->response_timeout = timeout;
	else
		s->response_timeout = CLT_TIMEOUT_MSEC;

	xstrncpy(s->name, name, sizeof(s->name));
	s->conn_type = conn_type;
	s->tout = timeout_new(hal_peer_event_default, s);
	s->state = HAL_CLT_STATE_DISCONN;
	s->side = side;

	if (s->side == HAL_CLT_RUNTIME)
		s->image_pos = -1;
	else
		s->image_pos = clt_num++;
	
	/* Add a new session to disconnected clients list. */
	list = slist_new();
	list->data = s;
	clt_disconn_list = slist_add_tail(clt_disconn_list, list);

	return s;
}

/* Add a new client. */
int hal_mgr_client_raw_new(const char *name, int wfd, int rfd,
			   int timeout, int side)
{
	struct hal_client *s;

	return_val_if_fail(name != NULL, -1);
	return_val_if_fail(wfd >= 0, -1);
	return_val_if_fail(rfd >= 0, -1);

	s = client_new(name, HAL_CONN_TYPE_RAW, timeout, side);
	/* Init peer base. */
	hal_peer_base_init(HAL_PEER_BASE(s), io, wfd, rfd, peer_event_handler);

	return 0;
}

int hal_mgr_client_unix_new(const char *name, const char *sock,
			    int timeout, int side)
{
	struct hal_client *s;
	int sfd;

	return_val_if_fail(name != NULL, -1);
	return_val_if_fail(sock != NULL, -1);

	sfd = sock_unix_connect(sock);
	if (sfd == -1)
		return -1;

	s = client_new(name, HAL_CONN_TYPE_UNIX, timeout, side);
	xstrncpy(s->sock, sock, sizeof(s->sock));
	/* Init peer base. */
	hal_peer_base_init(HAL_PEER_BASE(s), io, sfd, sfd, peer_event_handler);

	return 0;
}

int hal_mgr_client_inet_new(const char *name, const char *inet_addr,
			    int timeout, int side)
{
	struct hal_client *s;
	char *host, *port, *p;
	int sfd;

	return_val_if_fail(name != NULL, -1);

	p = strchr(inet_addr, ':');
	if (p == NULL)
		return -1;

	*p = 0;
	host = (char *)inet_addr;
	port = p+1;

	sfd = sock_inet_connect_wait("tcp", "ip", host, port, 0);
	if (sfd == -1)
		return -1;

	*p = ':';
	s = client_new(name, HAL_CONN_TYPE_INET, timeout, side);
	xstrncpy(s->inet_addr, inet_addr, sizeof(s->inet_addr));
	/* Init peer base. */
	hal_peer_base_init(HAL_PEER_BASE(s), io, sfd, sfd, peer_event_handler);

	return 0;
}

void hal_mgr_clients_reconnect()
{
	struct slist *iter;
	struct hal_client *s;
	char *host, *port, *p;
	int sfd;

	if (clt_error_list == NULL)
		return;

	/* All clients from the error list need to be reconnected. */
	for (iter = clt_error_list; iter != NULL; iter = iter->next) {
		s = (struct hal_client *)iter->data;

		assert(s->state == HAL_CLT_STATE_ERROR);
		assert(s->side != HAL_CLT_RUNTIME);

		/* Skip raw connects. */
		if (s->conn_type == HAL_CONN_TYPE_RAW)
			continue;

		switch (s->conn_type) {
		case HAL_CONN_TYPE_UNIX:
			sfd = sock_unix_connect(s->sock);
			break;
		case HAL_CONN_TYPE_INET:
			p = strchr(s->inet_addr, ':');
			if (p == NULL) {
				warn("wrong inet address %s", s->name);
				continue;
			}
			*p = 0;
			host = s->inet_addr;
			port = p+1;
			sfd = sock_inet_connect_wait("tcp", "ip", host, port, 0);
			*p = ':';
			break;
		default:
			break;
		}

		if (sfd == -1) {
			warn("can't connect to %s", s->name);
			continue;
		}

		hal_peer_base_init(HAL_PEER_BASE(s), io, sfd, sfd,
						peer_event_handler);
		/* Move a peer from the error list to the disconnected state
		 * list. */
		client_move(&clt_disconn_list, s);
	}

	/* Now try to connect with the hal logic. */
	hal_mgr_clients_connect();
}

/* Init manager internals. */
int hal_mgr_init(struct io_context *io_ptr)
{
	return_val_if_fail(io_ptr != NULL, -1);

	inputs_image = buffer_new();
	buffer_put_u8(inputs_image, 0xde);
	buffer_put_u8(inputs_image, 0xad);
	buffer_put_u8(inputs_image, 0xbe);
	buffer_put_u8(inputs_image, 0xaf);

	outputs_image = buffer_new();
	buffer_put_u8(outputs_image, 0xbe);
	buffer_put_u8(outputs_image, 0xaf);
	buffer_put_u8(outputs_image, 0xde);
	buffer_put_u8(outputs_image, 0xad);

	io = io_ptr;

	parser_init();

	return 0;
}

void static list_free(struct slist *head)
{
	struct slist *iter, *next;
	struct hal_client *s;

	iter = head;
	while (iter != NULL) {
		s = (struct hal_client *)iter->data;
		if (timeout_active(s->tout))
			timeout_del(s->tout);

		timeout_free(&s->tout);
		hal_peer_base_deinit(HAL_PEER_BASE(s));
		xfree(s);

		next = iter->next;
		slist_free(&iter);
		iter = next;
	}
}

/* Deinit manager and free resources. */
void hal_mgr_deinit()
{
	buffer_free(inputs_image);
	buffer_free(outputs_image);

	list_free(clt_conn_list);
	clt_conn_list = NULL;

	list_free(clt_disconn_list);
	clt_disconn_list = NULL;

	list_free(clt_error_list);
	clt_error_list = NULL;
}

